from __future__ import annotations

import logging
import random

import requests

from dynamicprompts.generators.promptgenerator import PromptGenerator, GeneratorException


logger = logging.getLogger(__name__)

class FeelingLuckyGenerator(PromptGenerator):
    def generate(self, search_query, num_prompts: int) -> list[str]:
        if search_query.strip() == "":
            query = random.randint(0, 10000000)
        else:
            query = search_query

        url = f"https://lexica.art/api/v1/search?q={query}"
        
        try:
            logger.info(f"Requesting {url}")
            response = requests.get(url)
            response.raise_for_status()
            data = response.json()
            images = data["images"]
            prompts = random.choices(images, k=num_prompts)
            return prompts
        except Exception as e:
            raise GeneratorException("Error while generating prompt: " + str(e))





from __future__ import annotations

import logging

from dynamicprompts.wildcardmanager import WildcardManager
from dynamicprompts import constants
from dynamicprompts.parser.combinatorial_generator import CombinatorialGenerator
from . import PromptGenerator

logger = logging.getLogger(__name__)


class CombinatorialPromptGenerator(PromptGenerator):
    def __init__(self, wildcard_manager: WildcardManager):
        self._wildcard_manager = wildcard_manager
        self._generator = CombinatorialGenerator(wildcard_manager)

    def generate(self, template, max_prompts=constants.MAX_IMAGES) -> list[str]:
        if template is None or len(template) == 0:
            return [""]
        prompts = self._generator.generate_prompts(template, max_prompts)
        prompts = list(prompts)

        return prompts



from __future__ import annotations

import logging
from random import Random
import random

from dynamicprompts import constants
from dynamicprompts.wildcardmanager import WildcardManager
from dynamicprompts.parser.random_generator import RandomGenerator

from . import PromptGenerator

logger = logging.getLogger(__name__)

class RandomPromptGenerator(PromptGenerator):
    def __init__(
        self,
        wildcard_manager: WildcardManager,
        seed: int | None = None,
        unlink_seed_from_prompt: bool = constants.UNLINK_SEED_FROM_PROMPT,
    ):
        self._wildcard_manager = wildcard_manager
        self._unlink_seed_from_prompt = unlink_seed_from_prompt

        if self._unlink_seed_from_prompt:
            self._random = random
        else:
            self._random = Random()
            if seed is not None:
                self._random.seed(seed)

        self._generator = RandomGenerator(wildcard_manager)

    def generate(self, template, max_prompts=constants.MAX_IMAGES) -> list[str]:
        if template is None or len(template) == 0:
            return [""]
        prompts = self._generator.generate_prompts(template, max_prompts)
        prompts = list(prompts)
        return prompts



from __future__ import annotations

from pathlib import Path


class WildcardFile:
    def __init__(self, path: Path, encoding="utf8"):
        self._path = path
        self._encoding = encoding

    def get_wildcards(self) -> set[str]:
        is_empty_line = (
            lambda line: line is None
            or line.strip() == ""
            or line.strip().startswith("#")
        )

        with self._path.open(encoding=self._encoding, errors="ignore") as f:
            lines = [line.strip() for line in f if not is_empty_line(line)]
            return set(lines)



from __future__ import annotations
from . import PromptGenerator
import random
import re
from tqdm import trange

from transformers import set_seed
from transformers import AutoTokenizer, AutoModelForCausalLM
from transformers import pipeline


MODEL_NAME = "Gustavosta/MagicPrompt-Stable-Diffusion"
MAX_SEED = 2 ** 32 - 1


class MagicPromptGenerator(PromptGenerator):
    generator = None

    def _load_pipeline(self):

        from modules.devices import get_optimal_device

        device = 0 if get_optimal_device() == "cuda" else -1

        if MagicPromptGenerator.generator is None:
            tokenizer = AutoTokenizer.from_pretrained(MODEL_NAME)
            model = AutoModelForCausalLM.from_pretrained(MODEL_NAME)

            MagicPromptGenerator.tokenizer = tokenizer
            MagicPromptGenerator.model = model
            MagicPromptGenerator.generator = pipeline(
                task="text-generation", tokenizer=tokenizer, model=model, device=device
            )

        return MagicPromptGenerator.generator

    def __init__(
        self,
        prompt_generator: PromptGenerator,
        max_prompt_length: int = 100,
        temperature: float = 0.7,
        seed: int | None = None,
    ):
        self._generator = self._load_pipeline()
        self._prompt_generator = prompt_generator
        self._max_prompt_length = max_prompt_length
        self._temperature = float(temperature)

        if seed is not None:
            set_seed(int(seed))

    def generate(self, *args, **kwargs) -> list[str]:
        prompts = self._prompt_generator.generate(*args, **kwargs)

        new_prompts = []
        for i in trange(len(prompts), desc="Generating Magic prompts"):
            orig_prompt = prompts[i]
            magic_prompt = self._generator(
                orig_prompt,
                max_length=self._max_prompt_length,
                temperature=self._temperature,
            )[0]["generated_text"]

            magic_prompt = self.clean_up_magic_prompt(orig_prompt, magic_prompt)
            new_prompts.append(magic_prompt)

        return new_prompts

    def clean_up_magic_prompt(self, orig_prompt, prompt):
        # remove the original prompt to keep it out of the MP fixes
        removed_prompt_prefix = False
        if re.search("^" + re.escape(orig_prompt), prompt):
            prompt = prompt.replace(orig_prompt, "", 1)
            removed_prompt_prefix = True

        # old-style weight elevation
        prompt = prompt.translate(str.maketrans("{}", "()")).strip()

        # useless non-word characters at the begin/end
        prompt = re.sub(r"^\W+|\W+$", "", prompt)

        # clean up whitespace in weighted parens
        prompt = re.sub(r"\(\s+", "(", prompt)
        prompt = re.sub(r"\s+\)", ")", prompt)

        # clean up whitespace in hyphens between words
        prompt = re.sub(r"\b\s+\-\s+\b", "-", prompt)
        prompt = re.sub(
            r"\s*[,;\.]+\s*(?=[a-zA-Z(])", ", ", prompt
        )  # other analogues to ', '
        prompt = re.sub(r"\s+_+\s+", " ", prompt)  # useless underscores between phrases
        prompt = re.sub(r"\b,\s*,\s*\b", ", ", prompt)  # empty phrases

        # Translate bangs into proper weight modifiers
        for match in re.findall(r"\b([\w\s\-]+)(\!+)", prompt):
            phrase = match[0]
            full_match = match[0] + match[1]
            weight = round(pow(1.1, len(match[1])), 2)

            prompt = prompt.replace(full_match, f"({phrase}:{weight})")

        # Put the original prompt back in
        if removed_prompt_prefix:
            prompt = orig_prompt + " " + prompt

        return prompt



from .promptgenerator import PromptGenerator
from .batched_combinatorial import BatchedCombinatorialPromptGenerator
from .combinatorial import CombinatorialPromptGenerator
from .dummygenerator import DummyGenerator
from .feelinglucky import FeelingLuckyGenerator
from .randomprompt import RandomPromptGenerator
from .attentiongenerator import AttentionGenerator



from __future__ import annotations

from typing import Iterable
from collections import OrderedDict

from .parse import (
    ActionBuilder,
    Parser,
)

from .commands import (
    Command,
    SequenceCommand,
    WildcardCommand,
    VariantCommand,
)

from dynamicprompts.wildcardmanager import WildcardManager


class CombinatorialSequenceCommand(SequenceCommand):
    def __init__(self, tokens: list[Command] | None = None, separator=" "):
        self._sep = separator
        super().__init__(tokens)

    def prompts(self, tokens: list[Command] | None = None) -> Iterable[str]:
        if tokens is None:
            tokens = self.tokens

        if len(tokens) == 0:
            yield ""
        else:
            token = tokens[0]
            for prompt in token.prompts():
                for next_prompts in self.prompts(tokens[1:]):
                    yield (prompt + self._sep + next_prompts).strip()


class CombinatorialWildcardCommand(WildcardCommand):
    def __init__(self, wildcard_manager, token):
        super().__init__(wildcard_manager, token)
        self._wildcard_manager = wildcard_manager
        self._wildcard = token[0]

    def prompts(self):
        generator = CombinatorialGenerator(self._wildcard_manager)
        values = self._wildcard_manager.get_all_values(self._wildcard)
        for val in values:
            for prompt in generator.generate_prompts(val):
                yield prompt

    def __repr__(self):
        return f"{self.__class__.__name__}({self._wildcard!r})"


class CombinatorialVariantCommand(VariantCommand):
    def _combo_to_prompt(self, combo: list[SequenceCommand]) -> Iterable[list[str]]:
        if len(combo) == 0:
            yield []
        else:
            c_1, c_rest = combo[0], combo[1:]

            for p in c_1.prompts():
                for rest_prompt in self._combo_to_prompt(c_rest):
                    if rest_prompt != "":
                        yield [p] + rest_prompt
                    else:
                        yield [p]

    def _dedupe(self, arr: list[str]) -> tuple[str]:
        d = OrderedDict()
        for item in arr:
            d[item] = None
        return tuple(d.keys())

    def prompts(self) -> Iterable[str]:
        if len(self._values) == 0:
            return []

        seen = set()

        for bound in range(self.min_bound, self.max_bound + 1):
            for combo in self._combinations(bound):
                for prompt_arr in self._combo_to_prompt(combo):
                    deduped_arr = self._dedupe(prompt_arr)
                    correct_size = len(deduped_arr) == bound
                    if deduped_arr not in seen and correct_size:
                        seen.add(deduped_arr)
                        yield self.sep.join(deduped_arr)

    def __repr__(self):
        z = zip(self._weights, self._values)
        return f"{self.__class__.__name__}({list(z)!r})"


class CombinatorialActionBuilder(ActionBuilder):
    def create_variant_command(self, variants, min_bound=1, max_bound=1, sep=","):
        return CombinatorialVariantCommand(variants, min_bound, max_bound, sep)

    def create_wildcard_command(self, token: str):
        return CombinatorialWildcardCommand(self._wildcard_manager, token)

    def create_sequence_command(self, token_list: list[Command]):
        return CombinatorialSequenceCommand(token_list)


class CombinatorialGenerator:
    def __init__(self, wildcard_manager):
        self._wildcard_manager = wildcard_manager

    def get_action_builder(self) -> ActionBuilder:
        return CombinatorialActionBuilder(self._wildcard_manager)

    def configure_parser(self) -> Parser:
        builder = self.get_action_builder()
        parser = Parser(builder)

        return parser

    def generate_prompts(
        self, prompt: str, num_prompts: int | None = None
    ) -> list[str]:
        if len(prompt) == 0:
            return []

        parser = self.configure_parser()
        sequence = parser.parse(prompt)
        prompts = sequence.prompts()

        if num_prompts is None:
            return [p for idx, p in enumerate(prompts)]
        else:
            return [p for idx, p in enumerate(prompts) if idx < num_prompts]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['dynamicprompts', 'dynamicprompts.generators', 'dynamicprompts.parser']

package_data = \
{'': ['*']}

install_requires = \
['pyparsing>=3.0.9,<4.0.0', 'requests>=2.28.1,<3.0.0', 'tqdm>=4.64.1,<5.0.0']

extras_require = \
{'attentiongrabber': ['spacy[attentiongrabber]>=3.4.4,<4.0.0']}

setup_kwargs = {
    'name': 'dynamicprompts',
    'version': '0.1.4',
    'description': 'Dynamic prompts templating library for Stable Diffusion',
    'long_description': '',
    'author': 'Adi Eyal',
    'author_email': 'adi@clearforest.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import logging
import random
from typing import cast, Iterable

from .parse import Parser, ActionBuilder
from .commands import (
    SequenceCommand,
    Command,
)
from dynamicprompts.wildcardmanager import WildcardManager

logger = logging.getLogger(__name__)


class RandomSequenceCommand(SequenceCommand):
    def __init__(self, tokens: list[Command] | None = None, separator=" "):
        self._sep = separator
        super().__init__(tokens)

    def prompts(self) -> Iterable[str]:
        if len(self.tokens) == 0:
            return []
        else:
            partials = [p.get_prompt() for p in self.tokens]
            return [self._sep.join(partials)]


class RandomWildcardCommand(Command):
    def __init__(self, wildcard_manager, token: str):
        super().__init__(token)
        self._wildcard_manager = wildcard_manager
        self._wildcard = token[0]

    def prompts(self) -> Iterable[str]:
        generator = RandomGenerator(self._wildcard_manager)
        values = self._wildcard_manager.get_all_values(self._wildcard)
        if len(values) == 0:
            logger.warning(f"No values found for wildcard {self._wildcard}")
            return [f"__{self._wildcard}__"]
        val = random.choice(values)
        prompts = generator.generate_prompts(val, 1)

        return prompts

    def __repr__(self):
        return f"{self.__class__.__name__}({self._wildcard!r})"


class RandomVariantCommand(Command):
    def __init__(self, variants, min_bound=1, max_bound=1, sep=","):
        super().__init__(variants)
        self._weights = [p["weight"][0] for p in variants]
        self._values = [p["val"] for p in variants]
        self.min_bound = min_bound
        self.max_bound = max_bound
        self.sep = sep
        self._remaining_values = self._values

    def _combo_to_prompt(self, combo: list[SequenceCommand]) -> Iterable[list[str]]:
        if len(combo) == 0:
            yield []
        else:
            c_1, c_rest = combo[0], combo[1:]

            for p in c_1.prompts():
                for rest_prompt in self._combo_to_prompt(c_rest):
                    if rest_prompt != "":
                        yield [p] + rest_prompt
                    else:
                        yield [p]

    def prompts(self) -> Iterable[str]:
        if len(self._values) == 0:
            return []

        num_choices = random.randint(self.min_bound, self.max_bound)
        combo = random.choices(self._values, weights=self._weights, k=num_choices)
        for prompt_arr in self._combo_to_prompt(combo):
            yield self.sep.join(prompt_arr)

    def __repr__(self):
        z = zip(self._weights, self._values)
        return f"{self.__class__.__name__}({list(z)!r})"


class RandomActionBuilder(ActionBuilder):
    def create_variant_command(self, variants, min_bound=1, max_bound=1, sep=","):
        return RandomVariantCommand(variants, min_bound, max_bound, sep)

    def create_wildcard_command(self, token: str):
        return RandomWildcardCommand(self._wildcard_manager, token)

    def create_sequence_command(self, token_list: list[Command]):
        return RandomSequenceCommand(token_list)


class RandomGenerator:
    def __init__(self, wildcard_manager: WildcardManager, rand=None):
        if rand is None:
            self._random = random
        else:
            self._random = rand
        self._wildcard_manager = wildcard_manager

    def get_action_builder(self) -> ActionBuilder:
        return RandomActionBuilder(self._wildcard_manager)

    def configure_parser(self):
        builder = self.get_action_builder()
        parser = Parser(builder)

        return parser.prompt

    def generate_prompts(self, prompt: str, num_prompts: int) -> list[str]:
        if len(prompt) == 0:
            return []

        parser = self.configure_parser()
        tokens = parser.parse_string(prompt)
        # tokens = cast(list[Command], tokens)

        squash_whitespace = lambda s: " ".join(s.split())

        generated_prompts = []
        for i in range(num_prompts):
            prompts = list(tokens[0].prompts())
            generated_prompts.append(squash_whitespace(prompts[0]))

        return generated_prompts
